import pickle

import keras
import numpy as np
from src.core.board import Board, GameState
from src.core.minimax import PExpNode
from src.core import minimax
from keras import losses
from keras.layers import Input, Convolution2D, Dense, Flatten, BatchNormalization
from keras.models import Model  # basic class for specifying and training a neural network
from sortedcontainers import SortedList


class PExpMind:
    def __init__(self, size, channels=1, init=True):

        self.size = size
        self.channels = channels

        if self.size == 7:
            self.value_est = self.value_model_7()
            self.policy_est = self.policy_model_7()
        elif self.size == 9:
            self.value_est = self.value_model_9()
            self.policy_est = self.policy_model_9()
        else:
            self.value_est = self.get_value_model()
            self.policy_est = self.get_policy_model()

        # search and training parameters
        self._max_expansion = lambda depth: np.inf if depth < 2 else 5
        self._p_threshold = lambda depth, log_p: log_p > -10
        self._move_convergence_count = lambda depth: 5

        self.p_exp_batch_size = 100
        self.max_iters = 10
        self.required_depth = 4

        # initialization
        init_examples = 10

        if init:
            sample_x = [np.random.randint(-1, 1, size=(init_examples, size, size, self.channels))]
            self.value_est.fit(sample_x, y=np.zeros(init_examples), epochs=1, batch_size=100)
            self.policy_est.fit(sample_x, y=np.zeros((init_examples, self.size ** 2)))

        self.train_vectors = []
        self.train_q = []
        self.train_p = []

        self.fitted = False

        self.memory_root = None

    def value_model_7(self):
        inp = Input(shape=(self.size, self.size, self.channels))

        # key difference between this and conv network is padding
        conv_1 = Convolution2D(64, (3, 3), padding='valid', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(inp)
        bn2 = BatchNormalization()(conv_1)
        conv_2 = Convolution2D(32, (3, 3), padding='valid', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn2)
        bn3 = BatchNormalization()(conv_2)

        flat = Flatten()(bn3)

        hidden = Dense(10, activation='relu', kernel_initializer='random_normal', use_bias=False)(flat)
        bn_final = BatchNormalization()(hidden)

        out = Dense(1, use_bias=False)(bn_final)

        model = Model(inputs=[inp], outputs=out)
        model.compile(loss=losses.mean_squared_error, optimizer='adam', metrics=['mean_squared_error'])

        return model

    def value_model_9(self):
        inp = Input(shape=(self.size, self.size, self.channels))

        # key difference between this and conv network is padding
        conv_1 = Convolution2D(64, (3, 3), padding='valid', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(inp)
        bn2 = BatchNormalization()(conv_1)
        conv_2 = Convolution2D(32, (3, 3), padding='valid', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn2)
        bn3 = BatchNormalization()(conv_2)
        conv_3 = Convolution2D(16, (3, 3), padding='valid', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn3)
        bn4 = BatchNormalization()(conv_3)

        flat = Flatten()(bn4)

        hidden = Dense(20, activation='relu', kernel_initializer='random_normal', use_bias=False)(flat)
        bn_final = BatchNormalization()(hidden)

        out = Dense(1, use_bias=False)(bn_final)

        model = Model(inputs=[inp], outputs=out)
        model.compile(loss=losses.mean_squared_error, optimizer='adam', metrics=['mean_squared_error'])

        return model

    def policy_model_7(self):
        inp = Input(shape=(self.size, self.size, self.channels))

        # key difference between this and conv network is padding
        conv_1 = Convolution2D(64, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(inp)
        bn2 = BatchNormalization()(conv_1)
        conv_2 = Convolution2D(32, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn2)
        bn3 = BatchNormalization()(conv_2)
        conv_3 = Convolution2D(16, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn3)
        bn4 = BatchNormalization()(conv_3)
        conv_4 = Convolution2D(8, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn4)
        bn5 = BatchNormalization()(conv_4)

        flat = Flatten()(bn5)

        hidden = Dense(self.size ** 2, activation='relu', kernel_initializer='random_normal', use_bias=False)(flat)
        bn_final = BatchNormalization()(hidden)

        out = Dense(self.size ** 2, activation='softmax')(bn_final)

        model = Model(inputs=[inp], outputs=out)
        model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

        return model


    def policy_model_9(self):
        inp = Input(shape=(self.size, self.size, self.channels))

        # key difference between this and value network
        conv_1 = Convolution2D(64, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(inp)
        bn2 = BatchNormalization()(conv_1)
        conv_2 = Convolution2D(64, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn2)
        bn3 = BatchNormalization()(conv_2)
        conv_3 = Convolution2D(64, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn3)
        bn4 = BatchNormalization()(conv_3)
        conv_4 = Convolution2D(32, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn4)
        bn5 = BatchNormalization()(conv_4)
        conv_5 = Convolution2D(16, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn5)
        bn6 = BatchNormalization()(conv_5)

        flat = Flatten()(bn6)

        hidden = Dense(self.size ** 2, activation='relu', kernel_initializer='random_normal', use_bias=False)(flat)
        bn_final = BatchNormalization()(hidden)

        out = Dense(self.size ** 2, activation='softmax')(bn_final)

        model = Model(inputs=[inp], outputs=out)
        model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

        return model

    def get_value_model(self):
        inp = Input(shape=(self.size, self.size, self.channels))

        # key difference between this and conv network is padding
        conv_1 = Convolution2D(64, (3, 3), padding='valid', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(inp)
        bn2 = BatchNormalization()(conv_1)
        conv_2 = Convolution2D(32, (3, 3), padding='valid', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn2)
        bn3 = BatchNormalization()(conv_2)
        #conv_3 = Convolution2D(32, (3, 3), padding='valid', activation='relu',
        #                       kernel_initializer='random_normal', use_bias=False)(bn3)
        #bn4 = BatchNormalization()(conv_3)
        #conv_4 = Convolution2D(16, (3, 3), padding='valid', activation='relu',
        #                       kernel_initializer='random_normal', use_bias=False)(bn4)
        #bn5 = BatchNormalization()(conv_4)

        flat = Flatten()(bn3)

        hidden = Dense(10, activation='relu', kernel_initializer='random_normal', use_bias=False)(flat)
        bn_final = BatchNormalization()(hidden)

        out = Dense(1, use_bias=False)(bn_final)

        model = Model(inputs=[inp], outputs=out)
        model.compile(loss=losses.mean_squared_error, optimizer='adam', metrics=['mean_squared_error'])

        return model

    def get_policy_model(self):
        inp = Input(shape=(self.size, self.size, self.channels))

        # key difference between this and conv network is padding
        conv_1 = Convolution2D(64, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(inp)
        bn2 = BatchNormalization()(conv_1)
        conv_2 = Convolution2D(32, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn2)
        bn3 = BatchNormalization()(conv_2)
        conv_3 = Convolution2D(16, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn3)
        bn4 = BatchNormalization()(conv_3)
        #conv_4 = Convolution2D(8, (3, 3), padding='same', activation='relu',
        #                       kernel_initializer='random_normal', use_bias=False)(bn4)
        #bn5 = BatchNormalization()(conv_4)

        flat = Flatten()(bn4)

        hidden = Dense(10, activation='relu', kernel_initializer='random_normal', use_bias=False)(flat)
        bn_final = BatchNormalization()(hidden)

        out = Dense(self.size ** 2, activation='softmax')(bn_final)

        model = Model(inputs=[inp], outputs=out)
        model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

        return model

    def pvs_k_principal_variations(self, leaf_nodes):
        # include the best move according to q
        principal_variations = []
        for item in leaf_nodes:
            if item.game_status == GameState.NOT_OVER:
                principal_variations.append(item)
            if len(principal_variations) == self.p_exp_batch_size:
                break
        return principal_variations

    def p_search(self, board, is_maximizing, root_node=None, save_root=True, consistency_check=False, verbose=True):

        if root_node is None:
            root_node = PExpNode(parent=None,
                                                is_maximizing=is_maximizing,
                                                full_move_list=minimax.MoveList(moves=(), position_hash=[]))

        principal_variations = [root_node]

        transposition_table = {}

        # all nodes at the leaves of the search tree
        leaf_nodes = SortedList(principal_variations, key=lambda x: x.p_comparator)

        i = 0
        MAXED_DURATION = 5

        previously_removed = set()

        principal_qs = []
        best_children = []
        # if we haven't completed enough iterations
        # or we're on a really shallow PV
        while True:
            i += 1

            # the ways P search terminates
            if i > self.max_iters:
                # break if there's some hanging issue
                if i > self.max_iters * 5:
                    break

                # same child has been best child for awhile now
                if len(set(best_children[-self._move_convergence_count:])) == 1\
                        and len(root_node.principal_variation.full_move_list) > self.required_depth:
                    break

                if root_node.principal_variation.game_status == GameState.DRAW:
                    break

            # searching doesn't get us anywhere for awhile
            if root_node.principal_variation and root_node.principal_variation.q:
                best_children.append(root_node.best_child)
                principal_qs.append(root_node.principal_variation.q)
                # had the same best q for MAXED_DURATION iterations
                if len(principal_qs) > MAXED_DURATION and np.abs(np.mean(principal_qs[-MAXED_DURATION : -1])) == 1:
                    if verbose:
                        print('Maxed Duration')
                    break

            # p expansion
            self.p_expand(board, principal_variations, leaf_nodes, transposition_table, previously_removed, verbose=verbose)

            principal_variations = self.pvs_k_principal_variations(leaf_nodes)

            # nothing left
            if not principal_variations:
                break

            # P will already have been expanded so do a Q eval
            #self.q_eval_top_leaves(leaf_nodes, min_eval_q=k ** 2, max_eval_q=k ** 2)
            self.q_eval(leaf_nodes)

            # this is used in case the p expansion doesn't capture the nodes we need to hit (if q_exp_batch_size is large enough this isn't needed)
            next_pvs = self.highest_leaf_qs(leaf_nodes, is_maximizing, max_p_eval=self.p_exp_batch_size * 2, num_leaves=self.q_exp_batch_size)
            print('Difference between P Expand and Q expand', len(next_pvs) + len(principal_variations), len(set(next_pvs + principal_variations)))

            #principal_variations.extend(next_pvs)

            if consistency_check:
                root_node.consistent_pv()

            # this check doesn't hold if we do PV Q extensions
            #if root_node.principal_variation:
            #    root_node.top_down_q()
            #    assert(abs(root_node.principal_variation.q - root_node.negamax()[0]) < 1E-4)

            # if we have a PV, add it to expand
            if root_node.principal_variation and root_node.principal_variation.game_status == GameState.NOT_OVER:
                # pv could point to a 'stable' node that doesn't expand
                #assert(root_node.principal_variation not in previously_removed)
                assert(not root_node.principal_variation.has_children())
                principal_variations.append(root_node.principal_variation)

            # remove duplicates
            principal_variations = set(principal_variations)

            print('Root', str(root_node))

        q_stats, p_stats = root_node.recursive_stats()
        print('Explored ' + str(p_stats) + " States (Q evals: " + str(q_stats) + ") in " + str(i) + " Iterations")

        possible_moves = root_node.get_sorted_moves()

        for move, q in possible_moves:
            node = root_node.children[move]
            print('Move', move, str(node))

        if save_root:
            self.pickle_root(board, root_node)

        return possible_moves, root_node


    def highest_leaf_qs(self, leaf_nodes, is_maximizing, max_p_eval=100, num_leaves=10):
        num_eval = min(max_p_eval, len(leaf_nodes))
        valid_leaves = [leaf for leaf in leaf_nodes.islice(0, num_eval) if leaf.is_assigned_q and leaf.game_status == GameState.NOT_OVER]
        best_leaves = sorted(valid_leaves, key=lambda x: abs(x.q), reverse=True)
        return best_leaves[:num_leaves]

    def q_eval(self, nodes):
        parents_to_update = set()
        board_matrices = []
        nodes = [node for node in nodes if not node.is_assigned_q]
        for leaf in nodes:
            # normally not, but it can be if nodes = PV's children
            assert(leaf.game_status == GameState.NOT_OVER)
            assert(not leaf.has_children())
            parents_to_update.update(leaf.parents)
            board_matrices.append(leaf.get_matrix())

        # if nothing to eval, get out
        if len(board_matrices) == 0:
            return

        # attach q predictions to all leaves and compute q tree at once
        q_predictions = np.clip(
                    self.value_est.predict(np.array(board_matrices), batch_size=1000).reshape(len(board_matrices)),
                    a_max=PExpNode.MAX_MODEL_Q,
                    a_min=PExpNode.MIN_MODEL_Q
            )

        for i, leaf in enumerate(nodes):
            # it's possible to have transposition assign q's even if we filtered above
            if not leaf.is_assigned_q:
                leaf.assign_q(q_predictions[i], GameState.NOT_OVER)

        for parent in parents_to_update:
            parent.recalculate_q()

    def q_eval_top_leaves(self, leaf_nodes, min_eval_q=10, max_eval_q=1000):
        # will only contain leaves of games not being over
        num_eval = min(max(min_eval_q, len(leaf_nodes)), max_eval_q)
        best_leaves = leaf_nodes.islice(0, num_eval)
        self.q_eval(best_leaves)

    # returns a list of board vectors for each node in nodes_to_expand
    def get_board_vectors(self, board, nodes_to_expand):
        p_search_vectors = []
        for parent in nodes_to_expand:
            # for each move except the last, make rapid moves on board
            if parent.has_matrix():
                p_search_vectors.append(parent.get_matrix())
            else:
                for move in parent.full_move_list.moves:
                    board.blind_move(*move)

                p_search_vectors.append(board.get_matrix())

                for _ in parent.full_move_list.moves:
                    board.unmove()
        return p_search_vectors

    accessed_transposition = 0

    # p_threshold(depth, p) returns a boolean vector filter
    # max_expansion(depth) returns an integer
    def define_policies(self, p_threshold, max_expansion, convergence_count,
                        alpha=0.2, p_exp_batch_size=25, q_exp_batch_size=25,
                        required_depth=4, max_iters=20):
        self._p_threshold = p_threshold
        self._max_expansion = max_expansion
        self._move_convergence_count = convergence_count

        self.p_exp_batch_size = p_exp_batch_size
        self.q_exp_batch_size = q_exp_batch_size
        self.alpha = alpha
        self.required_depth = required_depth
        self.max_iters = max_iters
    # Expand all nodes_to_expand, updating leaf_nodes in place
    # Also passes
    def p_expand(self, board, nodes_to_expand, leaf_nodes, transposition_table, previously_removed, verbose=True):
        for parent in nodes_to_expand:
            if parent in previously_removed:
                continue
            # parent could be a 'stable' node
            #assert(parent not in previously_removed)
            # should NOT be expanding any non-leaf node
            assert(parent in leaf_nodes)
            try:
                leaf_nodes.remove(parent)
            except ValueError:
                print(parent.p_comparator)
            previously_removed.add(parent)

        # each board state is defined by a list of moves
        p_search_vectors = self.get_board_vectors(board, nodes_to_expand)
        p_search_nodes = list(nodes_to_expand)

        p_board_vectors = np.array(p_search_vectors).reshape(len(p_search_vectors), self.size, self.size,
                                                             self.channels)

        log_p_predictions = np.log(self.policy_est.predict([p_board_vectors], batch_size=len(p_board_vectors))).reshape(-1)

        parent_index = np.hstack([np.full(shape=(self.size ** 2), fill_value=i) for i in range(len(nodes_to_expand))])
        parent_depth = np.hstack([np.full(shape=(self.size ** 2), fill_value=len(i.full_move_list.moves)) for i in nodes_to_expand])
        move_index = np.array([list(range(self.size ** 2)) * len(nodes_to_expand)])
        zipped_predictions = np.vstack([parent_index, parent_depth, move_index, log_p_predictions]).transpose()

        threshold_filter = self._p_threshold(zipped_predictions[:, 1], zipped_predictions[:, 3])
        filtered_predictions = zipped_predictions[threshold_filter]

        if verbose:
            print('Num P over threshold', len(filtered_predictions), 'out of', (len(nodes_to_expand) * self.size ** 2))

        new_leaves = []
        q_update = set()

        # numpy array of unique parent indices, start and end indices of each parent's move lists, and the count of the number of moves per parent
        unique_parents, move_indices, parent_counts = np.unique(filtered_predictions[:, 0], return_counts=True, return_index=True)

        # i is the parent meta index, index of parent index
        for i, parent_index in enumerate(unique_parents):
            parent = p_search_nodes[int(parent_index)]

            # build out the position really quickly (without checking if the game's over, since we know this is a valid position)
            for move in parent.full_move_list.moves:
                board.blind_move(*move)

            max_move_indices = move_indices[i + 1] if i < len(unique_parents) - 1 else filtered_predictions.shape[0]
            sorted_moves = sorted(filtered_predictions[move_indices[i]:max_move_indices, 2:4], key=lambda x: x[1], reverse=True)
            num_moves = min(len(sorted_moves), self._max_expansion(parent.depth()))

            self.create_children(board, parent, sorted_moves[:num_moves], new_leaves, q_update, transposition_table)

            for _ in parent.full_move_list.moves:
                board.unmove()

        for parent in q_update:
            parent.recalculate_q()

        leaf_nodes += new_leaves

        return len(new_leaves) > 0

    def create_children(self, board, parent, sorted_moves, new_leaves, q_update, transposition_table):
        for move_index, log_p_prediction in sorted_moves:
            child_move = self.index_to_move(move_index)
            if board.is_move_available(*child_move):
                child, made_new_child = parent.create_child(child_move, transposition_table)
                if made_new_child:
                    board.move(child_move[0], child_move[1])
                    child.assign_p(log_p_prediction)
                    # if game is over, then we have our q
                    if board.game_won():
                        winning_q = PExpNode.MIN_Q if board.get_player_to_move() == Board.FIRST_PLAYER else PExpNode.MAX_Q
                        child.assign_q(winning_q, GameState.WON)
                        q_update.add(parent)

                    elif board.game_assume_drawn():
                        child.assign_q(0, GameState.DRAW)
                        q_update.add(parent)

                    else:
                        new_leaves.append(child)
                        child.set_matrix(board.get_matrix())

                    # unmove for child
                    board.unmove()
                else:
                    PExpMind.accessed_transposition += 1
                    q_update.add(parent)

    def q(self, board):
        return self.value_est.predict([np.array([board.get_matrix().reshape(self.size, self.size, -1)])])[0][0]

    def pick_random_move(self, board, possible_moves):
        picked_action = 0

        # abs is only there to handle floating point problems
        qs = np.array([node.q for _, node in possible_moves])
        if board.get_player_to_move() == 1:
            distribution = np.abs(qs + 1.0) / 2
        else:
            # not sure this is correct
            distribution = sorted(-np.abs(qs - 1.0) / 2)

        if sum(distribution) > 0:
            distribution = (distribution.astype(np.float64) / sum(distribution))
            picked_action = np.random.choice(range(len(possible_moves)), 1, p=distribution)[0]

        return picked_action

    # used to pick a branch and fast forward it to use it as root node
    def cleanse_memory_tree(self, board, root_node, moves):
        if self.memory_root is not None:
            # self-play
            if np.alltrue(np.equal(self.memory_root.get_matrix(), board.get_matrix())):
                starting_root = self.memory_root
            for child in self.memory_root.children.values():
                if np.alltrue(np.equal(child.get_matrix(), board.get_matrix())):
                    #print('Found Previous Child')
                    starting_root = child
                    break
        #root_node.parents = []
        #root_node.cleanse_memory(moves)

        return None

    # with epsilon probability will select random move
    # returns whether game has concluded or not
    def make_move(self, board, as_player, verbose=True, epsilon=0.1, save_root=False, consistency_check=False):
        current_q = self.q(board)
        assert(as_player == board.get_player_to_move())

        # incomplete
        starting_root = self.cleanse_memory_tree(self.memory_root, None, None)

        # array of [best_move, best_node], root node of move calculations
        is_maximizing = True if board.get_player_to_move() == Board.FIRST_PLAYER else False
        possible_moves, root_node = self.p_search(board, is_maximizing,
                                                  root_node=starting_root,
                                                  consistency_check=consistency_check,
                                                  save_root=save_root, verbose=verbose)

        # best action is 0th index
        picked_action = 0

        # pick a suboptimal move randomly
        if np.random.random_sample() < epsilon:
            if verbose:
                print('suboptimal move')
            picked_action = self.pick_random_move(board, possible_moves)

        if len(possible_moves) == 0:
            print("Error in P Expand Search")

        picked_move, picked_node = possible_moves[picked_action]
        # add training example assuming best move
        best_move, best_node = possible_moves[0]
        best_q = best_node.q

        # q update with learning rate self.alpha
        #new_best_q = (1 - self.alpha) * current_q + self.alpha * best_q

        # compress to model valid range
        #new_best_q = max(min(new_best_q, PExpNode.MAX_MODEL_Q), PExpNode.MIN_MODEL_Q)

        #print(current_q, best_q)
        #self.add_train_example(board, new_best_q, best_move)

        # picked move may not equal best move if we're making a suboptimal one
        #board.move(*picked_move)

        self.memory_root = root_node.children[picked_move[0], picked_move[1]]

        return best_move, current_q, best_q

    def pickle_root(self, board, root_node):
        # save for debugging
        import sys
        sys.setrecursionlimit(100000)
        saving_hash = str(hash(tuple(board.get_matrix().reshape(-1))))
        with open('src/logs/' + saving_hash + '.pkl', 'wb') as f:
            root_node.set_matrix(board.get_matrix())
            pickle.dump(root_node, f)
            print('Root saved at ', saving_hash)

    def one_hot_p(self, move_index):
        vector = np.zeros((self.size ** 2))
        vector[move_index] = 1.
        return vector

    def move_to_index(self, move):
        return move[0] * self.size + move[1]

    def index_to_move(self, index):
        return np.int(index / self.size), np.int(index % self.size)

    # adds rotations
    def add_train_example(self, board, result, move):
        board_vectors = board.get_rotated_matrices()

        for i, vector in enumerate(board_vectors):
            self.train_vectors.append(vector)
            self.train_q.append(result)
            # get the i'th rotation
            which_rotation = board.get_rotated_point(self.move_to_index(move))[i]
            self.train_p.append(self.one_hot_p(which_rotation))


    def save(self, filename):
        self.value_est.save(filename + '_value.net')
        self.policy_est.save(filename + '_policy.net')

    def load_net(self, filename):
        self.value_est = keras.models.load_model(filename + '_value.net')
        self.policy_est = keras.models.load_model(filename + '_policy.net')
       
    def load(self, value_file, policy_file):
        self.value_est = keras.models.load_model(value_file)
        self.policy_est = keras.models.load_model(policy_file)


from collections import defaultdict

import numpy as np
import random
import json

from src.util import utils

class Move:
    def __init__(self, player, x, y):
        self.player = player
        self.x = x
        self.y = y


class GameState:
    WON = 1
    DRAW = 2
    NOT_OVER = 3

class BoardTransform:
    NUM_ROTATIONS = 8
    def __init__(self, size):
        self._size = size
        self._cached_point_rotations = defaultdict(list)
        self._cache_rotations()

    # Returns rotations and mirrors of the board state
    # This is important for teaching the convolution layers about rotational invariance
    def get_rotated_matrices(self, matrix):
        transposition_axes = (1, 0, 2)
        assert matrix.shape[0] == self._size
        return [
            matrix,
            np.transpose(matrix, axes=transposition_axes),
            np.rot90(matrix),
            np.transpose(np.rot90(matrix), axes=transposition_axes),
            np.rot90(matrix, 2),
            np.transpose(np.rot90(matrix, 2), axes=transposition_axes),
            np.rot90(matrix, 3),
            np.transpose(np.rot90(matrix, 3), axes=transposition_axes)
        ]

    def get_rotated_points(self, point):
        return self._cached_point_rotations[point]

    def coordinate_to_index(self, x, y):
        return x * self._size + y

    def index_to_coordinate(self, index):
        return int(index / self._size), int(index % self._size)

    # Precomputes some useful rotation values
    def _cache_rotations(self):
        indices = np.array(range(self._size ** 2)).reshape(self._size, self._size, 1)
        for i, matrix in enumerate(self.get_rotated_matrices(indices)):
            # iterate through the matrices
            for x in range(self._size):
                for y in range(self._size):
                    # where ever you see the matrix[x, y] value, indices[x, y] is where it originally started
                    self._cached_point_rotations[matrix[x, y, 0]].append(indices[x, y, 0])

# Board class represents the state of the game

# board perception will always be from the perspective of Player 1
# Q will always be from the perspective of Player 1 (Player 1 Wins = Q = 1, Player -1 Wins, Q = -1)

class Board:
    # channels
    NO_PLAYER = 0
    FIRST_PLAYER = 1
    SECOND_PLAYER = 2
    TURN_INFO_INDEX = 3

    STONE_PRESENT = 1
    STONE_ABSENT = 0

    def __init__(self, size=9, win_chain_length=5, draw_point=None):
        self._size = size

        # three for No Player, Player 1, Player 2, one for turn index
        self._matrix = np.zeros((self._size, self._size, 4), dtype=np.int)

        # Stone present for no player... make sense?
        self._matrix[:, :, Board.NO_PLAYER].fill(Board.STONE_PRESENT)

        # tracks which player played which spot (optimization)
        self._which_stone = np.zeros((self._size, self._size), dtype=np.int)

        self._win_chain_length = win_chain_length
        # stack of (x, y, player) tuple
        self._ops = []
        self._player_to_move = 1
        self._game_state = GameState.NOT_OVER
        self._available_moves = set()
        for i in range(self._size):
            for j in range(self._size):
                self._available_moves.add((i, j))

        self._num_moves = 0

        # number of plays at which it's a draw
        self.draw_point = draw_point if draw_point else self._size ** 2

    # for testing purposes
    def set_to_one_move_from_win(self):
        self.move(0, 0)
        self.move(0, 1)
        self.move(1, 0)
        self.move(1, 1)
        self.move(2, 0)
        self.move(2, 1)
        self.move(3, 0)
        self.move(3, 1)

    def unmove(self):
        previous_move = self._ops.pop()

        self._matrix[previous_move.x, previous_move.y, previous_move.player] = 0
        self._matrix[previous_move.x, previous_move.y, Board.NO_PLAYER] = 1
        self._which_stone[previous_move.x, previous_move.y] = Board.NO_PLAYER

        self._available_moves.add((previous_move.x, previous_move.y))
        self._flip_player_to_move()
        self._game_state = GameState.NOT_OVER

        self._num_moves -= 1

    def get_spot(self, x, y):
        return self._which_stone[x, y]

    def get_player_to_move(self):
        return self._player_to_move

    def get_player_last_move(self):
        return self._get_other_player(self.get_player_to_move())

    def get_last_move(self):
        return utils.peek_stack(self._ops)

    def get_size(self):
        return self._size

    # Places a stone at x, y for the next player's turn
    # Does not compute whether the game has completed or not (performance optimization)
    def blind_move(self, x, y):
        assert self._game_state is GameState.NOT_OVER
        self._ops.append(Move(self._player_to_move, x, y))

        self._matrix[x, y, self._player_to_move] = Board.STONE_PRESENT
        self._matrix[x, y, Board.NO_PLAYER] = Board.STONE_ABSENT

        self._which_stone[x, y] = self._player_to_move

        self._available_moves.remove((x, y))
        self._flip_player_to_move()

        self._num_moves += 1

    # Places a stone at x, y for the next player's turn
    # Computes whether game has concluded and if so, who the winner is
    def move(self, x, y):
        self.blind_move(x, y)
        self.compute_game_state()

    # Executes a random valid move
    def make_random_move(self):
        move_x, move_y = random.choice(list(self._available_moves))
        self.move(move_x, move_y)

    def _flip_player_to_move(self):
        self._player_to_move = self._get_other_player(self._player_to_move)

    def _get_other_player(self, player):
        if player == Board.FIRST_PLAYER:
            return Board.SECOND_PLAYER
        else:
            return Board.FIRST_PLAYER

    # returns None if game has not concluded, True if the last move won the game, False if draw
    # frequently called function, needs to be optimized
    def compute_game_state(self):
        last_move = utils.peek_stack(self._ops)
        if last_move:
            last_x, last_y = last_move.x, last_move.y
            if self.chain_length(last_x, last_y, -1, 0) + self.chain_length(last_x, last_y, 1, 0) >= self._win_chain_length + 1\
                    or self.chain_length(last_x, last_y, -1, 1) + self.chain_length(last_x, last_y, 1, -1) >= self._win_chain_length + 1\
                    or self.chain_length(last_x, last_y, 1, 1) + self.chain_length(last_x, last_y, -1, -1) >= self._win_chain_length + 1\
                    or self.chain_length(last_x, last_y, 0, 1) + self.chain_length(last_x, last_y, 0, -1) >= self._win_chain_length + 1:
                self._game_state = GameState.WON
                return
            if len(self._ops) == self._size ** 2:
                self._game_state = GameState.DRAW
                return
        self._game_state = GameState.NOT_OVER

    def _in_bounds(self, x, y):
        return 0 <= x < self._size and 0 <= y < self._size

    def chain_length(self, center_x, center_y, delta_x, delta_y):
        center_stone = self._which_stone[center_x, center_y]
        for chain_length in range(1, self._win_chain_length + 1):
            step_x = delta_x * chain_length
            step_y = delta_y * chain_length
            if not(0 <= center_x + step_x < self._size and 0 <= center_y + step_y < self._size and
                   self._matrix[center_x + step_x, center_y + step_y, center_stone] == 1):
                break
        return chain_length

    def is_move_available(self, x, y):
        return (x, y) in self._available_moves

    # runs a full compute
    def game_won(self):
        return self._game_state == GameState.WON

    # probably drawn, cheap check
    def game_assume_drawn(self):
        return len(self._ops) == self.draw_point

    def game_over(self):
        return self._game_state != GameState.NOT_OVER

    def get_winning_player(self):
        if self.game_over():
            return self._get_other_player(self._player_to_move)
        return Board.NO_PLAYER

    # simplest matrix representation of the board state
    def export(self):
        flatmatrix = np.zeros((self._size, self._size))
        for x in range(self._size):
            for y in range(self._size):
                if self._matrix[x, y, Board.FIRST_PLAYER] == 1:
                    flatmatrix[x, y] = Board.FIRST_PLAYER
                elif self._matrix[x, y, Board.SECOND_PLAYER] == 1:
                    flatmatrix[x, y] = Board.SECOND_PLAYER
        return {'size': str(self._size),
                           'win_chain_length': str(self._win_chain_length),
                            'boardstring': ''.join([str(elem) for elem in flatmatrix.astype(np.int32).reshape(-1)]),
                            'player_to_move': str(self._player_to_move)}

    # takes a board string and recreates a game state from it (quite slow)
    @classmethod
    def create_board_from_specs(Board, size, win_chain_length, boardstring):
        board = Board(size=size, win_chain_length=win_chain_length)
        player_1_moves = []
        player_2_moves = []
        reshaped_board = np.array([int(elem) for elem in boardstring]).reshape((size, size))
        for i in range(size):
            for j in range(size):
                if reshaped_board[i][j] == Board.FIRST_PLAYER:
                    player_1_moves.append((i, j))
                elif reshaped_board[i][j] == Board.SECOND_PLAYER:
                    player_2_moves.append((i, j))

        for i in range(len(player_1_moves)):
            board.blind_move(*player_1_moves[i])
            if i < len(player_2_moves):
                board.blind_move(*player_2_moves[i])

        return board

    @classmethod
    def load(Board, item_dict):
        return Board.create_board_from_specs(int(item_dict['size']),
                                             int(item_dict['win_chain_length']),
                                             item_dict['boardstring'])

    def pprint(self, lastmove_highlight=True):
        def display_char(x, y):
            move = utils.peek_stack(self._ops)
            if move:
                was_last_move = (x == move.x and y == move.y)
                if self._matrix[x, y, Board.FIRST_PLAYER] == 1:
                    if was_last_move and lastmove_highlight:
                        return 'X'
                    return 'x'
                elif self._matrix[x, y, Board.SECOND_PLAYER] == 1:
                    if was_last_move and lastmove_highlight:
                        return 'O'
                    return 'o'
            return ' '
        board_string = ""
        for i in range(0, self._size):
            board_string += "\n"
            for j in range(self._size):
                board_string += "|" + display_char(j, i)
            board_string += "|"
        return board_string

    def guide_print(self):
        def display_char(x, y):
            move = utils.peek_stack(self._ops)
            if move:
                was_last_move = (x == move.x and y == move.y)
                if self._matrix[x, y, Board.FIRST_PLAYER] == 1:
                    if was_last_move:
                        return 'X'
                    return 'x'
                elif self._matrix[x, y, Board.SECOND_PLAYER] == 1:
                    if was_last_move:
                        return 'O'
                    return 'o'
            return ' '
        board_string = " "

        for i in range(0, self._size):
            board_string += " " + str(i)
        for i in range(0, self._size):
            board_string += "\n" + str(i)
            for j in range(self._size):
                board_string += "|" + display_char(j, i)
            board_string += "|"
        return board_string

    def __str__(self):
        return self.pprint()





import numpy as np

from src.core.board import Board, BoardTransform
from src.core.game_record import GameRecord

from src.util import utils

class FeatureBoard_v1_1:
    CHANNELS = 4
    def __init__(self, board):
        self._size = board.get_size()
        self._ops = []
        self.tensor = np.zeros((self._size, self._size, FeatureBoard_v1_1.CHANNELS))
        self._player_to_move = FeatureBoard_v1_1._board_player_to_value(board.get_player_to_move())
        self._init(board)

    @staticmethod
    def _board_player_to_value(board_player):
        return 1 if board_player == Board.FIRST_PLAYER else -1

    def _init(self, board):
        for i in range(board.get_size()):
            for j in range(board.get_size()):
                if board.get_spot(i, j) == Board.FIRST_PLAYER:
                    self.tensor[i, j, 0] = 1
                elif board.get_spot(i, j) == Board.SECOND_PLAYER:
                    self.tensor[i, j, 1] = 1
        self._update_last_player()

    def _get_features(self):
        return np.copy(self.tensor)

    def get_p_features(self):
        return self._get_features()

    def get_q_features(self):
        return self._get_features()

    def _last_move(self):
        return utils.peek_stack(self._ops)

    def _clear_last_move(self, last_move):
        if last_move:
            self.tensor[last_move.x, last_move.y, 2] = 0

    def _update_last_move(self, last_move):
        if last_move:
            self.tensor[last_move.x, last_move.y, 2] = 1

    def _set_spot(self, move):
        player_index = 0 if self._player_to_move == Board.FIRST_PLAYER else 1
        self.tensor[move.x, move.y, player_index] = 1

    def _clear_spot(self, move):
        self.tensor[move.x, move.y, 0] = 0
        self.tensor[move.x, move.y, 1] = 0

    def _flip_player(self):
        if self._player_to_move == Board.FIRST_PLAYER:
            self._player_to_move = Board.SECOND_PLAYER
        else:
            self._player_to_move = Board.FIRST_PLAYER

    def _update_last_player(self):
        player_value = 1 if self._player_to_move == Board.FIRST_PLAYER else -1
        self.tensor[:, :, 3].fill(player_value)

    def move(self, move):
        assert move
        self._clear_last_move(self._last_move())
        self._ops.append(move)
        self._set_spot(move)
        self._update_last_move(move)
        self._update_last_player()
        self._flip_player()

    def unmove(self):
        last_move = self._ops.pop()
        self._clear_last_move(last_move)
        self._clear_spot(last_move)
        self._update_last_move(self._last_move())
        self._update_last_player()
        self._flip_player()

class FeatureSet_v1_1:
    CHANNELS = 4
    def __init__(self, record_string, learning_rate=0.2):
        self.q_features = []
        self.q_labels = []
        self.p_features = []
        self.p_labels = []
        self.channels = 4
        self._learning_rate = learning_rate
        self.iterate_on(GameRecord.parse(record_string))

    def make_feature_tensors(self, board, next_move, curr_q, next_q, make_p=True, make_q=True):
        # board's last move should be last_move, next_move not performed yet
        feature_tensor = FeatureBoard_v1_1(board).get_p_features()

        rot = BoardTransform(size=board.get_size())
        tensor_rotations = rot.get_rotated_matrices(feature_tensor)
        to_move_rotations = rot.get_rotated_points(rot.coordinate_to_index(*next_move))

        for i in range(len(tensor_rotations)):
            if make_q:
                self.q_features.append(tensor_rotations[i])
                # here is the q learning update
                self.q_labels.append(self.bound_q(curr_q * (1 - self._learning_rate) + next_q * self._learning_rate))
            if make_p:
                self.p_features.append(tensor_rotations[i])
                self.p_labels.append(to_move_rotations[i])

    def bound_q(self, q):
        return max(min(q, 1.0), -1.0)

    def get_q(self):
        return self.q_features, self.q_labels

    def get_p(self):
        return self.p_features, self.p_labels

    def iterate_on(self, record):
        board = Board.load(record.get_initial_state())
        q_assessments = record.get_q_assessments()

        for i, move in enumerate(record.get_moves()):
            if record.get_winning_player() == board.get_player_to_move():
                self.make_feature_tensors(board, move, q_assessments[i][0], q_assessments[i][1])
            # learn drawn positions
            elif record.get_winning_player() == Board.NO_PLAYER:
                self.make_feature_tensors(board, move, q_assessments[i][0], q_assessments[i][1], make_p=False)
            board.move(*move)


import numpy as np

from src.core.board import GameState

# search tree where we search strictly according to the likelihood function
class PExpNodeV2:
    MAX_Q = 1.0
    # largest Q allowed by model prediction (MAX_Q is a minimax certified win)
    MAX_MODEL_Q = 1.0 - 1E-4
    MIN_Q = -MAX_Q
    MIN_MODEL_Q = -MAX_MODEL_Q
    UNASSIGNED_Q = None

    def __init__(self, parent, is_maximizing, full_move_list):

        # with transposition, can have multiple parents
        if parent:
            self.parents = [parent]
        else:
            self.parents = []

        # full move tree
        self.full_move_list = full_move_list

        self.is_maximizing = is_maximizing
        # is the game over with this move list
        self.game_status = GameState.NOT_OVER
        # for debugging
        # key is move
        self.children = {}
        self.children_with_q = []
        # current best child
        self.best_child = None

        self.principal_variation = None
        self.q = PExpNodeV2.UNASSIGNED_Q

        # log likelihood of playing this move given root board state (used for PVS search)
        self.log_total_p = 0
        # log_total_p converted to integer that's easy to sort
        self.p_comparator = 0

        # only allowed to set once
        self.p_assigned = False

        # the quality of this move
        self.move_goodness = 0

        self._p_features = None
        self._q_features = None

    def set_q_features(self, q_features):
        self._q_features = q_features

    def set_p_features(self, p_features):
        self._p_features = p_features

    def get_q_features(self):
        return self._q_features

    def get_p_features(self):
        return self._p_features

    def has_q_features(self):
        return self._q_features is not None

    def has_p_features(self):
        return self._p_features is not None

    def has_children(self):
        return len(self.children) > 0

    transposition_access = 0

    # creates a new PExpNodeV2 and appends it as a child
    # will fish out of transposition table if it finds a suitable match
    # note this does NOT compute q for child
    def create_child(self, move, transposition_table):
        new_move_list = self.full_move_list.append(move)

        # build the hash of the transposition to check whether the position has been visited before
        transposition_hash = new_move_list.transposition_hash()
        child_found = transposition_hash in transposition_table
        child = None
        # transposed position exists
        if child_found:
            PExpNodeV2.transposition_access += 1
            child = transposition_table[transposition_hash]
            child.parents.append(self)
            if child.is_assigned_q():
                self.children_with_q.append(child)
        else:
            child = PExpNodeV2(parent=self,
                               is_maximizing=not self.is_maximizing,
                               full_move_list=new_move_list)
            transposition_table[transposition_hash] = child

        self.children[move] = child

        return child, not child_found

    def get_sorted_moves(self):
        valid_children = [tup for tup in self.children.items() if tup[1].q is not PExpNodeV2.UNASSIGNED_Q]
        return sorted(valid_children, key=lambda x: x[1].move_goodness, reverse=self.is_maximizing)

    # used ONLY for leaf assignment
    def assign_q(self, q, game_status):
        assert (len(self.children) == 0)

        # mark game status
        self.game_status = game_status

        # for a leaf node, the principal variation is itself
        assert not self.is_assigned_q()
        self.q = q
        assert(PExpNodeV2.MIN_Q <= q <= PExpNodeV2.MAX_Q)
        self.principal_variation = self

        for parent in self.parents:
            parent.children_with_q.append(self)

        self.assign_move_goodness()

    def assign_move_goodness(self):
        # play shorter sequences if advantageous, otherwise play longer sequences
        if self.q > 0:
            # because of the comparisons that use equality check in SortedSet, we need this value to be an integer
            # 1E6 multiplier allows the length of the full move list to be small in comparison to q
            self.move_goodness = int(self.q * 1E6 - len(self.principal_variation.full_move_list))
        else:
            self.move_goodness = int(self.q * 1E6 + len(self.principal_variation.full_move_list))

    # ab cutoff for when we know we can
    # Not sure if it applies regardless of parent
    def ab_valid(self):
        if len(self.parents) > 0 and self.parents[0].q:
            return self.parents[0].is_maximizing and self.parents[0].q < PExpNodeV2.MAX_Q \
                   or not self.parents[0].is_maximizing and self.parents[0].q > PExpNodeV2.MIN_Q
        return True

    def assign_p(self, log_p):
        assert(not self.p_assigned)
        self.log_total_p = self.parents[0].log_total_p + log_p
        self.p_assigned = True
        # Ideally
        self.p_comparator = int((min(-self.log_total_p, 100)) * 1E6)

    def depth(self):
        return len(self.full_move_list.moves)

    # returns if this node is better than another, comparing Q value
    def better_q(self, other):
        return self.move_goodness > other.move_goodness

    # Computes a bottom-up recomputation of Q
    def recalculate_q(self):
        # if this node is still a leaf, break
        if len(self.children_with_q) == 0:
            return

        # we are going to have to frequently search for the max, so might as well sort while we're at it
        sign = 1 if self.is_maximizing else -1
        self.children_with_q.sort(key=lambda x: sign * x.move_goodness, reverse=True)
        self.best_child = self.children_with_q[0]

        self.principal_variation = self.best_child.principal_variation
        self.q = self.best_child.q

        self.assign_move_goodness()

        # update parents
        for parent in self.parents:
            # if this node is in the PV line
            if parent.best_child == self or parent.best_child is None or (
                    (parent.is_maximizing and self.better_q(parent.best_child)) or
                    (not parent.is_maximizing and not self.better_q(parent.best_child))
            ):
                parent.recalculate_q()

    def cleanse_memory(self, moves):
        for child in self.children:
            child.cleanse_memory(moves)

    # DEBUGGING METHODS

    def is_assigned_q(self):
        return self.q != PExpNodeV2.UNASSIGNED_Q

    def recursive_stats(self):
        num_q = 1 if self.is_assigned_q() else 0
        num_nodes = len(self.children)
        if not self.has_children():
            self.num_nodes = num_nodes
            return num_q, num_nodes

        for child in self.children.values():
            child_q, child_nodes = child.recursive_stats()
            num_q += child_q
            num_nodes += child_nodes
        self.num_nodes = num_nodes
        return num_q, num_nodes

    def recursive_children(self, include_game_end=False):
        leaves = []
        for child in self.children.values():
            if child.has_children():
                leaves.extend(child.recursive_children())
            else:
                if not include_game_end and child.game_status == GameState.NOT_OVER\
                        or include_game_end:
                    leaves.append(child)
        return leaves

    # Assert consistency
    def consistent_pv(self):
        if self.has_children():
            valid_children = [tup for tup in self.children.items() if tup[1].q is not PExpNodeV2.UNASSIGNED_Q]
            assert(len(valid_children) == len(self.children_with_q))
            if len(valid_children) == 0:
                return

            if self.is_maximizing:
                best_move = max(valid_children, key=lambda x: x[1].move_goodness)[0]
            else:
                best_move = min(valid_children, key=lambda x: x[1].move_goodness)[0]

            assert(self.children[best_move].q - self.best_child.q < 1E-6)

            if self.is_assigned_q():
                if abs(self.q - self.children[best_move].q) > 1E-6:
                    print("Q Inconsistency")
                    print(self)
                    print(self.children[best_move])
                    self.principal_variation.recalculate_q()
                    self.children[best_move].principal_variation.recalculate_q()
                assert(abs(self.q - self.children[best_move].q) < 1E-6)
            assert(abs(self.principal_variation.q - self.children[best_move].principal_variation.q) < 1E-6)

            # using negamax framework
            #if self.principal_variation != self.children[best_move].principal_variation:
            #    self.whole_set_q()
            assert(self.principal_variation.log_total_p < 0)

            for child in self.children.values():
                child.consistent_pv()

    def top_down_q(self):
        valid_children = [tup for tup in self.children.items() if tup[1].q]
        if len(valid_children) == 0:
            return

        for _, child in valid_children:
            child.top_down_q()

        if self.is_maximizing:
            best_move = max(valid_children, key=lambda x: x[1].move_goodness)[0]
        else:
            best_move = min(valid_children, key=lambda x: x[1].move_goodness)[0]

        self.principal_variation = self.children[best_move].principal_variation
        self.q = self.children[best_move].q

    def negamax(self):
        valid_children = [tup for tup in self.children.items() if tup[1].q]
        if len(valid_children) == 0:
            return self.q, self

        best_child = None
        if self.is_maximizing:
            value = -np.inf
            for move, child in valid_children:
                child_value, next_child = child.negamax()
                assert(abs(child_value - next_child.principal_variation.q) < 1E-4)
                if child_value > value:
                    best_child = next_child
                    value = child_value
        else:
            value = np.inf
            for move, child in valid_children:
                child_value, next_child = child.negamax()
                assert (abs(child_value - next_child.principal_variation.q) < 1E-4)
                if child_value < value:
                    best_child = next_child
                    value = child_value

        return value, best_child

    # does formal expansion rather than a move list check
    def real_principal_variation(self):
        sign = 1 if self.is_maximizing else -1
        valid_children = [tup for tup in self.children.items() if tup[1].is_assigned_q()]
        if len(valid_children) == 0:
            return []
        best_move, best_child = max(valid_children, key=lambda x: sign * x[1].move_goodness)
        return [str(best_move)] + best_child.real_principal_variation()


    @classmethod
    # whether the q is from a game result (as opposed to an approximation of game state)
    def is_result_q(cls, q, epsilon=1E-7):
        return abs(q) > PExpNodeV2.MAX_Q - epsilon or abs(q) < epsilon

    def __str__(self):
        if self.principal_variation:
            return ("PV: " + ', '.join(self.real_principal_variation()) + " Q: {0:.4f} P: {1:.4f}").format(self.q, self.principal_variation.log_total_p)
        else:
            return "Position: " + str(self.full_move_list.moves) + " P: {0:.4f}".format(self.log_total_p)


import os
 
from setuptools import setup, find_packages 
 
 
setup(
    name="gomoku2",
    python_requires=">3",
    install_requires=[],
    include_package_data=True,
    packages=["core", 'main', 'learner', 'util'],
)



import pickle

import keras
import numpy as np
from src.core.board import Board, GameState
from src.learner.pexp_node_v2 import PExpNodeV2
from src.core import minimax
from keras import losses
from keras.layers import Input, Convolution2D, Dense, Flatten, BatchNormalization
from keras.models import Model  # basic class for specifying and training a neural network
from sortedcontainers import SortedList

from src.learner.game_to_features import FeatureSet_v1_1, FeatureBoard_v1_1

class PExpMind_v2:
    def __init__(self, size, init=True):

        self.size = size
        self.channels = FeatureSet_v1_1.CHANNELS

        assert size == 9
        self.value_est = self.value_model_9()
        self.policy_est = self.policy_model_9()

        # search and training parameters
        self._max_expansion = lambda depth: np.inf if depth < 2 else 5
        self._p_threshold = lambda depth, log_p: log_p > -10
        self._move_convergence_count = lambda depth: 5

        self.p_exp_batch_size = 100
        self.max_iters = 10
        self.required_depth = 4

        # initialization with random examples so we can immediately predict
        init_examples = 10

        if init:
            sample_x = [np.random.randint(-1, 1, size=(init_examples, size, size, self.channels))]
            self.value_est.fit(sample_x, y=np.zeros(init_examples), epochs=1, batch_size=100)
            self.policy_est.fit(sample_x, y=np.zeros((init_examples, self.size ** 2)))

        self.train_vectors = []
        self.train_q = []
        self.train_p = []

        self.fitted = False

        self.memory_root = None

    def value_model_9(self):
        inp = Input(shape=(self.size, self.size, self.channels))

        # key difference between this and conv network is padding
        conv_1 = Convolution2D(64, (3, 3), padding='valid', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(inp)
        bn2 = BatchNormalization()(conv_1)
        conv_2 = Convolution2D(32, (3, 3), padding='valid', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn2)
        bn3 = BatchNormalization()(conv_2)
        conv_3 = Convolution2D(16, (3, 3), padding='valid', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn3)
        bn4 = BatchNormalization()(conv_3)

        flat = Flatten()(bn4)

        hidden = Dense(20, activation='relu', kernel_initializer='random_normal', use_bias=False)(flat)
        bn_final = BatchNormalization()(hidden)

        out = Dense(1, use_bias=False)(bn_final)

        model = Model(inputs=[inp], outputs=out)
        model.compile(loss=losses.mean_squared_error, optimizer='adam', metrics=['mean_squared_error'])

        return model


    def policy_model_9(self):
        inp = Input(shape=(self.size, self.size, self.channels))

        # key difference between this and value network
        conv_1 = Convolution2D(64, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(inp)
        bn2 = BatchNormalization()(conv_1)
        conv_2 = Convolution2D(64, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn2)
        bn3 = BatchNormalization()(conv_2)
        conv_3 = Convolution2D(64, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn3)
        bn4 = BatchNormalization()(conv_3)
        conv_4 = Convolution2D(32, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn4)
        bn5 = BatchNormalization()(conv_4)
        conv_5 = Convolution2D(16, (3, 3), padding='same', activation='relu',
                               kernel_initializer='random_normal', use_bias=False)(bn5)
        bn6 = BatchNormalization()(conv_5)

        flat = Flatten()(bn6)

        hidden = Dense(self.size ** 2, activation='relu', kernel_initializer='random_normal', use_bias=False)(flat)
        bn_final = BatchNormalization()(hidden)

        out = Dense(self.size ** 2, activation='softmax')(bn_final)

        model = Model(inputs=[inp], outputs=out)
        model.compile(loss='categorical_crossentropy', optimizer='adam', metrics=['accuracy'])

        return model


    def pvs_k_principal_variations(self, leaf_nodes):
        # include the best move according to q
        principal_variations = []
        for item in leaf_nodes:
            if item.game_status == GameState.NOT_OVER:
                principal_variations.append(item)
            if len(principal_variations) == self.p_exp_batch_size:
                break
        return principal_variations

    class PEvenSearch:
        def __init__(self, board, max_iterations=10, verbose=True):
            self.board = board
            self.verbose = verbose
            self.is_maximizing = True if board.get_player_to_move() == Board.FIRST_PLAYER else False
            self.max_iterations = max_iterations

        def init(self):
            self.root_node = PExpNodeV2(parent=None,
                                      is_maximizing=self.is_maximizing,
                                      full_move_list=minimax.MoveList(moves=(), position_hash=[]))

            # transposition hash, Node
            self.transposition_table = {}


        def run(self):
            self.init()


    def p_search(self, board, is_maximizing, root_node=None, save_root=True, consistency_check=False, verbose=True):

        if root_node is None:
            root_node = PExpNodeV2(parent=None,
                                    is_maximizing=is_maximizing,
                                    full_move_list=minimax.MoveList(moves=(), position_hash=[]))

        principal_variations = [root_node]

        transposition_table = {}

        # all nodes at the leaves of the search tree
        leaf_nodes = SortedList(principal_variations, key=lambda x: x.p_comparator)

        i = 0
        MAXED_DURATION = 5

        previously_removed = set()

        principal_qs = []
        best_children = []
        # if we haven't completed enough iterations
        # or we're on a really shallow PV
        while True:
            i += 1

            # the ways P search terminates
            if i > self.max_iters:
                break
                # break if there's some hanging issue
                #if i > self.max_iters * 5:
                #    break

                # same child has been best child for awhile now
                #if len(set(best_children[-self._move_convergence_count:])) == 1\
                #        and len(root_node.principal_variation.full_move_list) > self.required_depth:
                #    break

                #if root_node.principal_variation.game_status == GameState.DRAW:
                #    break

            # searching doesn't get us anywhere for awhile
            if root_node.principal_variation and root_node.principal_variation.q:
                best_children.append(root_node.best_child)
                principal_qs.append(root_node.principal_variation.q)
                # had the same best q for MAXED_DURATION iterations
                if len(principal_qs) > MAXED_DURATION and np.abs(np.mean(principal_qs[-MAXED_DURATION : -1])) == 1:
                    if verbose:
                        print('Maxed Duration')
                    break

            # p expansion
            self.p_expand(board, principal_variations, leaf_nodes, transposition_table, previously_removed, verbose=verbose)

            principal_variations = self.pvs_k_principal_variations(leaf_nodes)

            # nothing left
            if not principal_variations:
                break

            # P will already have been expanded so do a Q eval
            #self.q_eval_top_leaves(leaf_nodes, min_eval_q=k ** 2, max_eval_q=k ** 2)
            self.q_eval(leaf_nodes)

            # this is used in case the p expansion doesn't capture the nodes we need to hit (if q_exp_batch_size is large enough this isn't needed)
            next_pvs = self.highest_leaf_qs(leaf_nodes, is_maximizing, max_p_eval=self.p_exp_batch_size * 2, num_leaves=self.q_exp_batch_size)
            print('Difference between P Expand and Q expand', len(next_pvs) + len(principal_variations), len(set(next_pvs + principal_variations)))

            #principal_variations.extend(next_pvs)

            if consistency_check:
                root_node.consistent_pv()

            # this check doesn't hold if we do PV Q extensions
            #if root_node.principal_variation:
            #    root_node.top_down_q()
            #    assert(abs(root_node.principal_variation.q - root_node.negamax()[0]) < 1E-4)

            # if we have a PV, add it to expand
            if root_node.principal_variation and root_node.principal_variation.game_status == GameState.NOT_OVER:
                # pv could point to a 'stable' node that doesn't expand
                #assert(root_node.principal_variation not in previously_removed)
                assert(not root_node.principal_variation.has_children())
                principal_variations.append(root_node.principal_variation)

            # remove duplicates
            principal_variations = set(principal_variations)

            print('Root', str(root_node))

        q_stats, p_stats = root_node.recursive_stats()
        print('Explored ' + str(p_stats) + " States (Q evals: " + str(q_stats) + ") in " + str(i) + " Iterations")

        possible_moves = root_node.get_sorted_moves()

        for move, q in possible_moves:
            node = root_node.children[move]
            print('Move', move, str(node))

        if save_root:
            self.pickle_root(board, root_node)

        return possible_moves, root_node


    def highest_leaf_qs(self, leaf_nodes, is_maximizing, max_p_eval=100, num_leaves=10):
        num_eval = min(max_p_eval, len(leaf_nodes))
        valid_leaves = [leaf for leaf in leaf_nodes.islice(0, num_eval) if leaf.is_assigned_q() and leaf.game_status == GameState.NOT_OVER]
        best_leaves = sorted(valid_leaves, key=lambda x: abs(x.q), reverse=True)
        return best_leaves[:num_leaves]

    def q_eval(self, nodes):
        parents_to_update = set()
        board_matrices = []
        nodes = [node for node in nodes if not node.is_assigned_q()]
        for leaf in nodes:
            # normally not, but it can be if nodes = PV's children
            assert(leaf.game_status == GameState.NOT_OVER)
            assert(not leaf.has_children())
            parents_to_update.update(leaf.parents)
            board_matrices.append(leaf.get_q_features())


        # if nothing to eval, get out
        if len(board_matrices) == 0:
            return

        # attach q predictions to all leaves and compute q tree at once
        q_predictions = np.clip(
                    self.value_est.predict(np.array(board_matrices), batch_size=1000).reshape(len(board_matrices)),
                    a_max=PExpNodeV2.MAX_MODEL_Q,
                    a_min=PExpNodeV2.MIN_MODEL_Q
            )

        for i, leaf in enumerate(nodes):
            # it's possible to have transposition assign q's even if we filtered above
            if not leaf.is_assigned_q():
                leaf.assign_q(q_predictions[i], GameState.NOT_OVER)

        for parent in parents_to_update:
            parent.recalculate_q()

    def q_eval_top_leaves(self, leaf_nodes, min_eval_q=10, max_eval_q=1000):
        # will only contain leaves of games not being over
        num_eval = min(max(min_eval_q, len(leaf_nodes)), max_eval_q)
        best_leaves = leaf_nodes.islice(0, num_eval)
        self.q_eval(best_leaves)

    # returns a list of board vectors for each node in nodes_to_expand
    def get_board_vectors(self, board, nodes_to_expand):
        p_search_vectors = []
        for parent in nodes_to_expand:
            # for each move except the last, make rapid moves on board
            if parent.has_p_features():
                p_search_vectors.append(parent.get_p_features())
            else:
                for move in parent.full_move_list.moves:
                    self.feature_board.move(move)

                features = self.feature_board.get_p_features()

                p_search_vectors.append(features)

                for _ in parent.full_move_list.moves:
                    self.feature_board.unmove()
        return p_search_vectors

    accessed_transposition = 0

    # p_threshold(depth, p) returns a boolean vector filter
    # max_expansion(depth) returns an integer
    def define_policies(self, p_threshold, max_expansion, convergence_count,
                        alpha=0.2, p_exp_batch_size=25, q_exp_batch_size=25,
                        required_depth=4, max_iters=20):
        self._p_threshold = p_threshold
        self._max_expansion = max_expansion
        self._move_convergence_count = convergence_count

        self.p_exp_batch_size = p_exp_batch_size
        self.q_exp_batch_size = q_exp_batch_size
        self.alpha = alpha
        self.required_depth = required_depth
        self.max_iters = max_iters
    # Expand all nodes_to_expand, updating leaf_nodes in place
    # Also passes
    def p_expand(self, board, nodes_to_expand, leaf_nodes, transposition_table, previously_removed, verbose=True):
        for parent in nodes_to_expand:
            if parent in previously_removed:
                continue
            # parent could be a 'stable' node
            #assert(parent not in previously_removed)
            # should NOT be expanding any non-leaf node
            assert(parent in leaf_nodes)
            try:
                leaf_nodes.remove(parent)
            except ValueError:
                print(parent.p_comparator)
            previously_removed.add(parent)

        # each board state is defined by a list of moves
        p_search_vectors = self.get_board_vectors(board, nodes_to_expand)
        p_search_nodes = list(nodes_to_expand)

        p_board_vectors = np.array(p_search_vectors).reshape((len(p_search_vectors), self.size, self.size, self.channels))

        log_p_predictions = np.log(self.policy_est.predict([p_board_vectors], batch_size=len(p_board_vectors))).reshape(-1)

        parent_index = np.hstack([np.full(shape=(self.size ** 2), fill_value=i) for i in range(len(nodes_to_expand))])
        parent_depth = np.hstack([np.full(shape=(self.size ** 2), fill_value=len(i.full_move_list.moves)) for i in nodes_to_expand])
        move_index = np.array([list(range(self.size ** 2)) * len(nodes_to_expand)])
        zipped_predictions = np.vstack([parent_index, parent_depth, move_index, log_p_predictions]).transpose()

        threshold_filter = self._p_threshold(zipped_predictions[:, 1], zipped_predictions[:, 3])
        filtered_predictions = zipped_predictions[threshold_filter]

        if verbose:
            print('Num P over threshold', len(filtered_predictions), 'out of', (len(nodes_to_expand) * self.size ** 2))

        new_leaves = []
        q_update = set()

        # numpy array of unique parent indices, start and end indices of each parent's move lists, and the count of the number of moves per parent
        unique_parents, move_indices, parent_counts = np.unique(filtered_predictions[:, 0], return_counts=True, return_index=True)

        # i is the parent meta index, index of parent index
        for i, parent_index in enumerate(unique_parents):
            parent = p_search_nodes[int(parent_index)]

            # build out the position really quickly (without checking if the game's over, since we know this is a valid position)
            for move in parent.full_move_list.moves:
                board.blind_move(*move)

            max_move_indices = move_indices[i + 1] if i < len(unique_parents) - 1 else filtered_predictions.shape[0]
            sorted_moves = sorted(filtered_predictions[move_indices[i]:max_move_indices, 2:4], key=lambda x: x[1], reverse=True)
            num_moves = min(len(sorted_moves), self._max_expansion(parent.depth()))

            self.create_children(board, parent, sorted_moves[:num_moves], new_leaves, q_update, transposition_table)

            for _ in parent.full_move_list.moves:
                board.unmove()

        for parent in q_update:
            parent.recalculate_q()

        leaf_nodes += new_leaves

        return len(new_leaves) > 0

    def create_children(self, board, parent, sorted_moves, new_leaves, q_update, transposition_table):
        for move_index, log_p_prediction in sorted_moves:
            child_move = self.index_to_move(move_index)
            if board.is_move_available(*child_move):
                child, made_new_child = parent.create_child(child_move, transposition_table)
                if made_new_child:
                    board.move(child_move[0], child_move[1])
                    child.assign_p(log_p_prediction)
                    # if game is over, then we have our q
                    if board.game_won():
                        winning_q = PExpNodeV2.MIN_Q if board.get_player_to_move() == Board.FIRST_PLAYER else PExpNodeV2.MAX_Q
                        child.assign_q(winning_q, GameState.WON)
                        q_update.add(parent)

                    elif board.game_assume_drawn():
                        child.assign_q(0, GameState.DRAW)
                        q_update.add(parent)

                    else:
                        new_leaves.append(child)
                        child.set_p_features(self.feature_board.get_p_features())
                        child.set_q_features(self.feature_board.get_q_features())

                    # unmove for child
                    board.unmove()
                else:
                    PExpMind_v2.accessed_transposition += 1
                    q_update.add(parent)

    def q(self):
        q_features = self.feature_board.get_q_features()
        return self.value_est.predict([np.array([q_features])])[0][0]

    def pick_random_move(self, board, possible_moves):
        picked_action = 0

        # abs is only there to handle floating point problems
        qs = np.array([node.q for _, node in possible_moves])
        if board.get_player_to_move() == 1:
            distribution = np.abs(qs + 1.0) / 2
        else:
            # not sure this is correct
            distribution = sorted(-np.abs(qs - 1.0) / 2)

        if sum(distribution) > 0:
            distribution = (distribution.astype(np.float64) / sum(distribution))
            picked_action = np.random.choice(range(len(possible_moves)), 1, p=distribution)[0]

        return picked_action

    # used to pick a branch and fast forward it to use it as root node
    def cleanse_memory_tree(self, board, root_node, moves):
        if self.memory_root is not None:
            # self-play
            if np.alltrue(np.equal(self.memory_root.get_matrix(), board.get_matrix())):
                starting_root = self.memory_root
            for child in self.memory_root.children.values():
                if np.alltrue(np.equal(child.get_matrix(), board.get_matrix())):
                    #print('Found Previous Child')
                    starting_root = child
                    break
        #root_node.parents = []
        #root_node.cleanse_memory(moves)

        return None




    # with epsilon probability will select random move
    # returns whether game has concluded or not
    def make_move(self, board, as_player, verbose=True, epsilon=0.1, save_root=False, consistency_check=False):

        self.feature_board = FeatureBoard_v1_1(board)
        current_q = self.q()
        assert(as_player == board.get_player_to_move())

        # incomplete
        starting_root = self.cleanse_memory_tree(self.memory_root, None, None)

        # array of [best_move, best_node], root node of move calculations
        is_maximizing = True if board.get_player_to_move() == Board.FIRST_PLAYER else False
        possible_moves, root_node = self.p_search(board, is_maximizing,
                                                  root_node=starting_root,
                                                  consistency_check=consistency_check,
                                                  save_root=save_root, verbose=verbose)

        # best action is 0th index
        picked_action = 0

        # pick a suboptimal move randomly
        #if np.random.random_sample() < epsilon:
        #    if verbose:
        #        print('suboptimal move')
        #    picked_action = self.pick_random_move(board, possible_moves)

        if len(possible_moves) == 0:
            print("Error in P Expand Search")

        picked_move, picked_node = possible_moves[picked_action]
        # add training example assuming best move
        best_move, best_node = possible_moves[0]
        best_q = best_node.q

        print(current_q, best_q)

        return best_move, current_q, best_q

    def pickle_root(self, board, root_node):
        # save for debugging
        import sys
        sys.setrecursionlimit(100000)
        saving_hash = str(hash(tuple(board.get_matrix().reshape(-1))))
        with open('src/logs/' + saving_hash + '.pkl', 'wb') as f:
            root_node.set_matrix(board.get_matrix())
            pickle.dump(root_node, f)
            print('Root saved at ', saving_hash)

    def one_hot_p(self, move_index):
        vector = np.zeros((self.size ** 2))
        vector[move_index] = 1.
        return vector

    def move_to_index(self, move):
        return move[0] * self.size + move[1]

    def index_to_move(self, index):
        return np.int(index / self.size), np.int(index % self.size)


    def save(self, filename):
        self.value_est.save(filename + '_value.net')
        self.policy_est.save(filename + '_policy.net')

    def load_net(self, filename):
        self.value_est = keras.models.load_model(filename + '_value.net')
        self.policy_est = keras.models.load_model(filename + '_policy.net')
       
    def load(self, value_file, policy_file):
        self.value_est = keras.models.load_model(value_file)
        self.policy_est = keras.models.load_model(policy_file)



import numpy as np
from src.core.board import Board
from src.core.game_record import GameRecord

# a match between two ai's
class Match:
    def __init__(self, mind1, mind2, size=9, random_seed=42, opening_moves=10, draw_point=50, trivialize=False):
        self.board = Board(size=size, win_chain_length=5, draw_point=draw_point)

        random_state = np.random.RandomState(random_seed)

        if not trivialize:
            # randomize the board a bit
            for j in range(random_state.randint(0, opening_moves)):
                self.board.make_random_move()
        else:
            self.board.set_to_one_move_from_win()

        self.players = {Board.FIRST_PLAYER: mind1, Board.SECOND_PLAYER: mind2}
        self.game_record = GameRecord.create(self.board)

    def play(self):
        print(self.board)
        current_player = self.board.get_player_to_move()

        while True:
            move, current_q, best_q = self.players[current_player].make_move(self.board,
                                                    as_player=current_player,
                                                    epsilon=0.1,
                                                    consistency_check=False,
                                                    verbose=True)
            print(self.board.pprint())

            self.game_record.add_move(move, current_q, best_q)
            self.board.move(*move)

            if self.board.game_over():
                self.game_record.set_winner(self.board.get_winning_player())
                break

            if current_player == Board.FIRST_PLAYER:
                current_player = Board.SECOND_PLAYER
            else:
                current_player = Board.FIRST_PLAYER

        return self.game_record.export()




from src.learner import pexp_mind
from src.core.board import Board
from src.core.minimax import PExpNode
import numpy as np

import random

SIZE = 9
CHANNELS = 4

import os
# bug with python on macos
os.environ['KMP_DUPLICATE_LIB_OK']='True'

# good luck!
if __name__ == "__main__":
    mind = pexp_mind.PExpMind(size=SIZE, init=False, channels=CHANNELS)
    mind.load_net('../trained_models/9_4_4')

    def expanding_p(depth, p):
        return np.logical_or.reduce([
            np.logical_and(depth < 2, p > -6),
            np.logical_and(depth < 4, p > -4),
            np.logical_and(depth < 6, p > -4),
            np.logical_and(depth < np.inf, p > -3)
        ])


    def permissive_expansion(depth):
        if depth < 2:
            return np.inf
        if depth < 8:
            return 5
        return 3


    mind.define_policies(expanding_p, permissive_expansion, convergence_count=5,
                         alpha=0.2, q_exp_batch_size=SIZE ** 2,
                         p_exp_batch_size=SIZE ** 3, required_depth=6, max_iters=25)

    board = Board(size=SIZE, win_chain_length=5)

    # randomize the board a bit
    for j in range(random.randint(0, int(SIZE * 2.5))):
        board.make_random_move()

    print(board.guide_print())

    while True:
        if board.get_player_to_move() == Board.FIRST_PLAYER:
            inp = input("Input your move (i.e. \"3 5\"): ")
            if len(inp.split(' ')) != 2:
                print('Incorrect number of coordinates, please try again!')
                continue
            x, y = inp.split(' ')
            try:
                x = int(x)
                y = int(y)
            except:
                print('Please input Numbers!')
                continue
            if x < 0 or x >= SIZE or y < 0 or y >= SIZE:
                print('Out of bounds!')
                continue
            if not board.is_move_available(x, y):
                print('Invalid Move!')
                continue
            result = board.move(x, y)
            print(board.guide_print())
        else:
            print('Computer is thinking...')

            possible_moves, root_node = mind.p_search(board, False, root_node=None, save_root=False)

            best_move, best_node = possible_moves[0]
            print(" ")
            print(best_move, 'Q:', best_node.q)

            if best_node.q > PExpNode.MAX_MODEL_Q:
                print('Computer Resigns!')
                break

            board.move(*best_move)
            print(board.guide_print())

        if board.game_approx_drawn():
            print("DRAW!")
            break

        if board.game_won():
            if board.get_player_to_move() == 1:
                print('COMPUTER WINS!')
            else:
                print('YOU WIN!')
            break



import numpy as np
import json
import pickle

class GameRecord:

    # winning_player values
    DRAW = 0
    # winning_player = Board.FIRST_PLAYER or Board.SECOND_PLAYER
    def __init__(self, initial_state_string, moves, winning_player, q_assessments):
        self.initial_state = initial_state_string
        self.moves = moves
        self._winning_player = winning_player
        self.q_assessments = q_assessments

    # initial_board_state (Board): initial position of the game
    @classmethod
    def create(cls, initial_board_state):
        return GameRecord(initial_board_state.export(), [], GameRecord.DRAW, [])

    @classmethod
    def parse(cls, string):
        obj_dict = json.loads(string)
        return GameRecord(obj_dict['initial_state'],
                          obj_dict['moves'],
                          obj_dict['winning_player'],
                          obj_dict['q_assessments'])

    class PrettyFloat(float):
        def __repr__(self):
            return '%.15g' % self

    def add_move(self, move, curr_q, best_q):
        # assert we have not concluded the game yet
        if self._winning_player != GameRecord.DRAW:
            raise ValueError
        self.moves.append(move)
        self.q_assessments.append((GameRecord.PrettyFloat(curr_q), GameRecord.PrettyFloat(best_q)))

    def set_winner(self, player):
        self._winning_player = player

    def export(self):
        obj_dict = {
            'initial_state': self.initial_state,
            'moves': self.moves,
            'winning_player': self._winning_player,
            'q_assessments': self.q_assessments
        }
        return json.dumps(obj_dict)

    def get_initial_state(self):
        return self.initial_state

    def get_moves(self):
        return self.moves

    def get_winning_player(self):
        return self._winning_player

    def get_q_assessments(self):
        return self.q_assessments